from typing import Optional
from typing import Tuple

import pandas as pd
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.data import random_split

from ..datasets.binary_image_classification_ds import BinaryImageClassificationDS


class BinaryImageClassificationDM(LightningDataModule):

    """
    Example of LightningDataModule for MNIST dataset.

    A DataModule implements 5 key methods:
        - prepare_data (things to do on 1 GPU/TPU, not on every GPU/TPU in distributed mode)
        - setup (things to do on every accelerator in distributed mode)
        - train_dataloader (the training dataloader)
        - val_dataloader (the validation dataloader(s))
        - test_dataloader (the test dataloader(s))

    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data

    Read the docs:
        https://pytorch-lightning.readthedocs.io/en/latest/extensions/datamodules.html
    """

    def __init__(
        self,
        data_dir: str,
        scaling: int,
        num_classes: int,
        train_val_test_split: Tuple[float, float, float] = (0.8, 0.1, 0.1),
        batch_size: int = 64,
        num_workers: int = 0,
        pin_memory: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.data_df = pd.read_csv(data_dir, index_col=0)
        self.file_paths = self.data_df["file_path"].tolist()
        self.labels = self.data_df["label"].tolist()
        self.scaling = scaling
        self.train_val_test_split = train_val_test_split
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.num_classes = num_classes

        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None

    def prepare_data(self):
        """Download data if needed. This method is called only from a single GPU.
        Do not use it to assign state (self.x = y)."""
        pass

    def setup(self, stage: Optional[str] = None):
        """Load data. Set variables: self.data_train, self.data_val, self.data_test."""

        # print(paths)

        dataset = BinaryImageClassificationDS(
            self.file_paths, self.labels, self.scaling
        )
        total = len(dataset)
        n_train = int(total * self.train_val_test_split[0])
        n_val = int(total * self.train_val_test_split[1])
        n_test = int(total * self.train_val_test_split[2])

        # print(total, n_train, n_va)

        while (n_train + n_val + n_test) < total:
            n_train += 1
        while (n_train + n_val + n_test) > total:
            n_train -= 1
        self.data_train, self.data_val, self.data_test = random_split(
            dataset, (n_train, n_val, n_test)
        )

        print("len train", len(self.data_train))
        print("len val", len(self.data_val))
        print("len test", len(self.data_test))

    def train_dataloader(self):
        return DataLoader(
            dataset=self.data_train,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=True,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.data_val,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.data_test,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )



# from pymongo.collection import Collection
import warnings
from typing import List
from pathlib import Path
from pymongo.collection import Collection
import os
from bson.objectid import ObjectId
import cv2
import pandas as pd


def get_prediction_task_df(
    label: str,
    version: float,
    db_images: Collection,
    origin: str = None,
    limit: int = 1000,
    predict_annotated: bool = False,
) -> pd.DataFrame:
    """Function queries the given collection for images to predict the given label on.\
        Images which already have an annotation of this label or a prediction with the current\
        AI version are excluded. If given a origin, only images of the given origin are predicted.\
        Returns a pd.DataFrame with columns "_id" (image id in database) and "file_path" (complete filepath on our server)

    Args:
        label (str): label to query for
        version (float): AI version number
        db_images (Collection): collection containing images
        origin (str, optional): Optionally filter for a specific origin. Defaults to None.
        limit (int, optional): Number to limit the output. Defaults to 1000.

    Returns:
        pd.DataFrame: Dataframe with columns "_id" and "file_path"
    """
    if origin:
        agg = [{"$match": {"origin": origin}}]
    else:
        agg = []

    agg.extend(
        get_images_to_prelabel_query(
            label=label,
            version=version,
            limit=limit,
            predict_annotated=predict_annotated,
        )
    )
    result = db_images.aggregate(agg)

    df_dict = {"file_path": [], "_id": []}
    for _ in result:
        df_dict["_id"].append(_["_id"])
        df_dict["file_path"].append(_["path"])

    return pd.DataFrame().from_dict(df_dict)


def get_binary_train_df(
    label: str, neg_label_list: List[str], db_images
) -> pd.DataFrame:
    """Queries db_images for pos. and neg. images of given label / neg label list

    Args:
        label (str): label defined as positive-
        neg_label_list (List[(str, float)]): List of tuples containing labelname and\
            multiplier. Multiplier is used to limit entries of each label to n_pos * multiplier
        db_images (): pymongo collection instance

    Returns:
        pd.DataFrame: pd.Dataframe containing columns "file_path" and "label". Label has value\
            0 or 1.
    """
    # Positive Images
    agg = [{"$match": {f"labels_new.{label}": True}}]
    imgs_pos = [_["path"] for _ in db_images.aggregate(agg)]
    n_pos = len(imgs_pos)
    imgs_pos_df = {"file_path": [_ for _ in imgs_pos], "label": [1 for _ in imgs_pos]}

    # Negative Images
    imgs_neg = [_["path"] for _ in db_images.find({f"labels_new.{label}": False})]

    for neg_label, multiplier in neg_label_list:
        agg = [
            {"$match": {f"labels_new.{neg_label}": True}},
            {"$limit": n_pos * multiplier},
        ]
        _imgs_neg = [_["path"] for _ in db_images.aggregate(agg)]
        imgs_neg.extend(_imgs_neg)

    imgs_neg_df = {"file_path": [_ for _ in imgs_neg], "label": [0 for _ in imgs_neg]}

    df_dict = imgs_pos_df
    df_dict["file_path"].extend(imgs_neg_df["file_path"])
    df_dict["label"].extend(imgs_neg_df["label"])

    label_df = pd.DataFrame.from_dict(df_dict)
    return label_df


def get_image_db_template(
    origin: str = None,
    intervention_id: ObjectId = None,
    path: str = None,
    n_frame: int = None,
):
    """Returns db_images template

    Args:
        origin (str, optional): String to define source. Defaults to None.
        intervention_id (ObjectId, optional): Id of intervention the image originates from. Defaults to None.
        path (str, optional): path to image. Defaults to None.
        n_frame (int, optional): Integer reffering to the frame number in the original video. Defaults to None.

    Returns:
        dict
    """

    template = {
        "origin": origin,
        "intervention_id": intervention_id,
        "path": path,
        "n_frame": n_frame,
        "in_progress": False,
    }
    return template


def extract_frames_to_db(
    frames: List[int],
    intervention_id: ObjectId,
    base_path_frames: Path,
    db_images,
    db_interventions,
    verbose=False,
):
    """Extracts given frames (as List of integers) from intervention with given id.

    Args:
        frames (List[int]): List of integers referring to frame numbers.
        intervention_id (ObjectId): Intervention id to extract from
        base_path_frames (Path): Base Path to store frames in
        db_images ([type]): pymongo collection instance
        db_interventions ([type]): pymongo collection instance
        verbose (bool, optional): Triggers console output. Defaults to False.

    Returns:
        List[ObjectId]: Returns list of inserted ids in db_images
    """
    intervention = db_interventions.find_one({"_id": intervention_id})
    video_key = intervention["video_key"]
    frame_directory = base_path_frames.joinpath(video_key)
    assert frame_directory.exists()
    cap = cv2.VideoCapture(intervention["video_path"])
    if "frames" in intervention:
        intervention_frame_dict = intervention["frames"]
    else:
        intervention_frame_dict = {}
    inserted_image_ids = []

    for n_frame in frames:
        if str(n_frame) in intervention_frame_dict:
            if verbose:
                print(f"Frame {n_frame} was already extracted, passing")
            continue
        frame_path = frame_directory.joinpath(f"{n_frame}.png")
        cap.set(cv2.CAP_PROP_POS_FRAMES, n_frame)
        success, image = cap.read()
        assert success
        cv2.imwrite(frame_path.as_posix(), image)

        db_image_entry = get_image_db_template(
            origin=intervention["origin"],
            intervention_id=intervention["_id"],
            path=frame_path.as_posix(),
            n_frame=n_frame,
        )

        _id = db_images.insert_one(db_image_entry).inserted_id
        intervention_frame_dict[str(n_frame)] = _id
        db_interventions.update_one(
            {"_id": intervention["_id"]}, {"$set": {"frames": intervention_frame_dict}}
        )
        inserted_image_ids.append(_id)
        if verbose:
            print(db_image_entry)

    return inserted_image_ids


def delete_frames_from_db(
    image_ids: List[ObjectId], db_images, db_interventions, verbose: bool = False
):
    """Function expect List of image ids. Deletes them in db_images and also deletes their reference in\
        db_interventions. Only works with extracted frames from videos!

    Args:
        image_ids (List[ObjectId]): List of Ids for frames to delete.
        db_images ([type]): pymongo collection instance. Entries must have fields "intervention_id" and "n_frame"
        db_interventions ([type]): pymongo collection instance. Entries must have field "frames"\
            containing a dictionary with frame number as keys.
        verbose (bool, optional): Enables console output. Defaults to False.
    """
    for image_id in image_ids:
        image = db_images.find_one({"_id": image_id})
        n_frame = image["n_frame"]
        intervention_id = image["intervention_id"]
        db_interventions.update_one(
            {"_id": intervention_id}, {"$unset": {f"frames.{n_frame}": ""}}
        )
        db_images.delete_one({"_id": image_id})
        if verbose:
            print(image)
            print("deleted\n\n")


def get_images_to_prelabel_query(
    label: str, version: int, predict_annotated: bool = False, limit: int = 100000
):
    """Function expects a label (str) and returns a query to find images which
    have no prediction or an outdated prediction for this label.

    Args:
        label (str): label to query for
        version (float): float value representing the current prelabel AI version
        limit (int, optional): Maximum value of images to return. Defaults to 100000.

    Returns:
        List: List of dictionaries containing the elements for a pymongo query aggregation
    """
    return [
        {
            "$match": {
                "$and": [
                    {
                        "$or": [
                            {f"predictions.{label}": {"$exists": False}},
                            {f"predictions.{label}.version": {"$lt": version}},
                        ]
                    },
                    {f"labels.annotation.{label}": {"$exists": predict_annotated}},
                ]
            }
        },
        {"$limit": limit},
    ]


def get_images_for_tasks(
    label: str,
    upper_confidence_threshold: float,
    lower_confidence_threshold: float,
    inclusive: bool,
    limit: int,
    db_images,
):
    """Queries db_images for images to annotate. Images will be selected\
        between confidence threshholds (inclusive True) or outside confidence\
        threshholds (inclusive False).\
        Images with labels_unclear.unclear == True will be excluded.\
        Only images with highest ai-version will be included.


    Args:
        label (str): label to query for
        upper_confidence_threshold (float): value between 0 and 1.
        lower_confidence_threshold (float): value between 0 and 1.
        inclusive (bool): If True, included images have prediction values\
            between threshholds. Otherwise, these images are excluded
        limit (int): Maximum amount of images to return
        db_images ([type]): pymongo collection instance.

    Returns:
        [type]: [description]
    """
    latest_ai_version = max(db_images.distinct(f"predictions.{label}.version"))
    if not inclusive:
        confidence_aggregation = {
            "$or": [
                {f"predictions.{label}.value": {"$gt": upper_confidence_threshold}},
                {f"predictions.{label}.value": {"$lt": lower_confidence_threshold}},
            ]
        }
    else:
        confidence_aggregation = {
            "$and": [
                {f"predictions.{label}.value": {"$lt": upper_confidence_threshold}},
                {f"predictions.{label}.value": {"$gt": lower_confidence_threshold}},
            ]
        }

    _images = db_images.aggregate(
        [
            {
                "$match": {
                    "$and": [
                        {f"labels_new.{label}": {"$exists": False}},  # Unlabeled Images
                        {
                            "$or": [
                                {
                                    "labels_unclear.unclear": False
                                },  # Images have not been labeled as unclear
                                {"labels_unclear.unclear": {"$exists": False}},
                            ]
                        },
                        {f"predictions.{label}.version": latest_ai_version},
                        confidence_aggregation,
                    ]
                }
            },
            {"$sample": {"size": limit}},
        ]
    )

    _ids = []
    images = []
    for image in _images:
        if image["_id"] not in _ids:
            _ids.append(image["_id"])
            images.append(image)

    return images


def exctract_frame_list(
    video_key: str, frame_list: List[int], base_path_frames: Path, db_interventions: str
):
    """Function to extract frames.

    Args:
        video_key (str): [description]
        frame_list (List[int]): [description]
        base_path_frames (Path): [description]
        db_interventions (str): [description]
    """
    intervention = db_interventions.find_one({"video_key": video_key})
    assert base_path_frames.exists()

    if not intervention:
        warnings.warn(f"Intervention with video_key {video_key} does not exist")

    frames_path = base_path_frames.joinpath(video_key)

    if not frames_path.exists():
        os.mkdir(frames_path)

    video_path = Path(intervention["video_path"])
    cap = cv2.VideoCapture(video_path.as_posix())

    for n_frame in frame_list:
        cap.set(cv2.CAP_PROP_POS_FRAMES, n_frame)
        ret, frame = cap.read()
        if ret:
            cv2.imwrite(frames_path.joinpath(f"{n_frame}.png"), frame)


def get_images_in_progress(db_images):
    return db_images.find({"in_progress": True})



from typing import List
from .binary_image_classification_ds import BinaryImageClassificationDS
from torch.utils.data import Dataset


def load_binary_image_classification_dataset(paths: List, ids: List, scaling: int, training: bool) -> Dataset:
    dataset = BinaryImageClassificationDS(paths, ids, scaling=scaling, training=training)

    return dataset



from typing import List
from bson.objectid import ObjectId
import torch
from datetime import datetime as dt


def report_to_string(report: dict[str]) -> str:
    """Function to merge the elements of the report field to a single string element.

    Args:
        report (dict): Dict with report categories as keys and lists of strings as values

    Returns:
        str: Concatenated string of the report.
    """
    report_string = []
    for key, value in report.items():
        _report_string = "\n".join(value)
        _report_string = "\n".join([key, _report_string])
        _report_string += "\n"
        report_string.append(_report_string)

    report_string = "\n".join(report_string)
    return report_string


def binary_classification_pred_to_db_value(
    pred: torch.Tensor, ids: List, version: float, targets: List, creation_date=dt.now()
):
    """Function expects tensor of a predicted batch and list of corresponding database ids.
    Appplies sigmoid function and assigns True or False values at a threshold of >= 0.5.

    Args:
        pred (torch.Tensor): Batch of tensor. Number of predictions for each image must match length of targets.
        ids (List): List of ObjectIds matching the images in database. Length must match length of batch.
        version (float): Version of the according prelabel algorithm.
        targets (List): string list of labels which were predicted. Order and length must match predictions of network
        creation_date ([type], optional): Datetime of prediction. Defaults to dt.now().

    Returns:
        [dict]: dictionary with ObjectIds as keys and according db values as labels.
        Should be inserted at "labels.predictions.{prelabel_engine}
    """
    pred = torch.sigmoid(pred).cpu().numpy()
    assert len(pred[0]) is len(targets)
    bool_pred = pred.copy()
    bool_pred[bool_pred >= 0.5] = 1
    bool_pred[bool_pred < 0.5] = 0

    values = {
        ObjectId(ids[i]): {
            "version": version,
            "targets": targets,
            "value": float(pred[i][0]),
            "creation_date": creation_date,
            "label": bool(bool_pred[i][0]),
        }
        for i in range(len(bool_pred))
    }

    return values



import warnings
from pathlib import Path
from .pl_ileum_detection_resnet import IleumDetectionResnet
from .pl_tool_detection_resnet import ToolDetectionResnet
from .pl_polyp_detection_resnet import PolypDetectionResnet
import torch

AVAILABLE_AI_MODELS = ["polyp", "ileum", "tool", "ileocaecalvalve", "appendix"]


def get_ckpt_path(ai_name: str, version: float, path_models: Path) -> Path:
    """Function selects folder which matches the label in the given folder and then\
        returns a Path object of the corresponding model version

    Args:
        ai_name (str): name of the ai
        version (float): version number (one decimal!)
        path_models (Path): Path object pointing to the base model folder

    Returns:
        Path: Path object pointing to the desired model
    """

    version = "{:.1f}".format(version)
    _path = path_models.joinpath(f"{ai_name}/{version}.ckpt")
    if not _path.exists():
        warnings.warn(f"Path {_path} does not exist!")
        print(_path)

    return _path


def load_model(model_name: str, version: float, _eval: bool, base_path_models: Path):
    """Function loads model for given name.

    Args:
        model_name (str): One of "polyp", "ileum", "tool", "ileocaecalvalve", "appendix"
        version (float): version number.
        eval (bool): If True, model is loaded in eval mode.
        base_path_models(Path): Points to directory containing folders named by models, containing checkpoints named by
            version numbers (e.g. '0.1.ckpt')

    Returns:
        [PL Model]: Model for given name.
    """
    assert model_name in AVAILABLE_AI_MODELS

    ckpt_path = get_ckpt_path(model_name, version, base_path_models)
    if model_name == "polyp":
        Model = PolypDetectionResnet
    if model_name == "ileum":
        Model = IleumDetectionResnet
    if model_name == "tool":
        Model = ToolDetectionResnet
    if model_name == "ileocaecalvalve":
        Model = IleumDetectionResnet
    if model_name == "appendix":
        Model = IleumDetectionResnet

    trained_model = Model.load_from_checkpoint(checkpoint_path=ckpt_path)

    if torch.cuda.is_available():
        trained_model.cuda(0)
    # trained_model.to(0)
    if _eval:
        trained_model.train(False)
        trained_model.freeze()

    return trained_model



from typing import Any
from typing import List

import torch
import torch.nn as nn
from pytorch_lightning import LightningModule
from torch import sigmoid
from torchmetrics.classification.accuracy import Accuracy
from torchvision import models


class ToolDetectionResnet(LightningModule):
    def __init__(self, num_classes, freeze_extractor, **kwargs):
        super().__init__()
        self.num_classes = num_classes
        self.freeze_extractor = freeze_extractor

        # this line ensures params passed to LightningModule will be saved to ckpt
        # it also allows to access params with 'self.hparams' attribute
        self.save_hyperparameters()
        self.model = models.resnext50_32x4d(pretrained=True)
        print(self.model)

        if self.freeze_extractor:
            print("Transfer learning with a fixed ConvNet feature extractor")
            for param in self.model.parameters():
                param.requires_grad = False
        else:
            print("Transfer learning with a full ConvNet finetuning")

        num_ftrs = self.model.fc.in_features
        self.model.fc = nn.Linear(num_ftrs, self.num_classes)

        # loss function
        self.criterion = nn.BCEWithLogitsLoss()  # nn.NLLLoss

        self.train_accuracy = Accuracy()
        self.val_accuracy = Accuracy()
        self.test_accuracy = Accuracy()

    def forward(self, x: torch.Tensor):
        return self.model(x)

    def step(self, batch: Any):
        inputs, labels = batch
        outputs = self.forward(inputs)
        preds = sigmoid(outputs).squeeze()
        preds[preds >= 0.5] = 1
        preds[preds < 0.5] = 0
        loss = self.criterion(outputs.squeeze(), torch.tensor(labels).type_as(outputs))

        return loss, preds, labels

    def training_step(self, batch: Any, batch_idx: int):
        loss, preds, targets = self.step(batch)

        acc = self.train_accuracy(preds, targets)

        # log train metrics
        self.log("train/loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("train/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        # and then read it in some callback or in training_epoch_end() below
        # remember to always return loss from training_step, or else backpropagation will fail!
        return {"loss": loss, "preds": preds, "targets": targets}

    def training_epoch_end(self, outputs: List[Any]):
        # `outputs` is a list of dicts returned from `training_step()`
        pass

    def validation_step(self, batch: Any, batch_idx: int):
        loss, preds, targets = self.step(batch)

        # log val metrics
        acc = self.val_accuracy(preds, targets)
        self.log("val/loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("val/acc", acc, on_step=False, on_epoch=True, prog_bar=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def validation_epoch_end(self, outputs: List[Any]):
        pass

    def test_step(self, batch: Any, batch_idx: int):
        loss, preds, targets = self.step(batch)

        # log test metrics
        acc = self.test_accuracy(preds, targets)
        self.log("test/loss", loss, on_step=False, on_epoch=True)
        self.log("test/acc", acc, on_step=False, on_epoch=True)

        return {"loss": loss, "preds": preds, "targets": targets}

    def test_epoch_end(self, outputs: List[Any]):
        pass

    def configure_optimizers(self):
        """Choose what optimizers and learning-rate schedulers to use in your optimization.
        Normally you'd need one. But in the case of GANs or similar you might have multiple.

        See examples here:
            https://pytorch-lightning.readthedocs.io/en/latest/common/lightning_module.html#configure-optimizers
        """
        optimizer = torch.optim.Adam(
            params=self.parameters(),
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )

        lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            mode="min",
            factor=0.1,
            patience=3,
            threshold=0.0001,
            threshold_mode="rel",
            cooldown=0,
            min_lr=0,
            eps=1e-08,
            verbose=False,
        )

        return {
            "optimizer": optimizer,
            "lr_scheduler": lr_scheduler,
            "monitor": "val/loss",
        }



import albumentations as A
import cv2
import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision import transforms

img_augmentations = A.Compose(
    [
        A.HorizontalFlip(p=0.5),
        A.MotionBlur(p=0.3, blur_limit=(10, 20)),
        A.RandomBrightnessContrast(p=0.5),
    ]
)

cropping_large = A.Compose([A.CenterCrop(width=1024, height=1024)])

cropping_small = A.Compose([A.CenterCrop(width=720, height=720)])

img_transforms = transforms.Compose([transforms.ToTensor()])


class BinaryImageClassificationDS(Dataset):
    def __init__(self, paths, labels, scaling: int = 75, training: bool = True):
        self.paths = paths
        self.scaling = scaling
        self.training = training

        self.labels = labels
        assert len(paths) == len(labels)

        self.classes = {0: "negative", 1: "positive"}

    def __getitem__(self, idx):
        # img = cv2.cvtColor(cv2.imread(self.paths[idx]), cv2.COLOR_BGR2RGB)
        img = cv2.imread(self.paths[idx])
        width = int(1024 * self.scaling / 100)  # img.shape[1]
        height = int(1024 * self.scaling / 100)  # img.shape[0]

        if img.shape[1] >= 1024 and img.shape[0] >= 1024:
            img = cropping_large(image=img)["image"]
        elif img.shape[1] >= 720 and img.shape[0] >= 720:
            img = cropping_small(image=img)["image"]
        dim = (width, height)
        # Switch BGR to RGB
        img = np.flip(img, axis=-1)
        if self.training:
            img = img_augmentations(image=img)["image"]

        img = cv2.resize(img, dsize=dim, interpolation=cv2.INTER_AREA)
        img = cv2.normalize(
            img, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
        if self.training:
            return img_transforms(img), torch.tensor(self.labels[idx])
        else:
            return img_transforms(img), self.labels[idx]

    def __len__(self):
        return len(self.paths)



from collections import defaultdict
import json
from typing import List


def get_default_dict():
    """Helper function returns a defaultdict generating dictionaries"""
    return defaultdict(dict)


def get_label_count_by_origin(label_list: List[str], db_images):
    count_all = defaultdict(get_default_dict)
    for label in label_list:
        r = db_images.aggregate(
            [
                {
                    "$group": {
                        "_id": {"label": f"$labels_new.{label}", "origin": "$origin"},
                        "count": {"$sum": 1},
                    }
                }
            ]
        )

        r = [_ for _ in r if "origin" in _["_id"] and "label" in _["_id"]]
        # r.sort()
        for category_count in r:
            count_all[label][category_count["_id"]["origin"]][
                category_count["_id"]["label"]
            ] = category_count["count"]

    return count_all


def get_label_count(label_list: List[str], db_images):
    count_all = defaultdict(dict)
    for label in label_list:
        r = db_images.aggregate(
            [{"$group": {"_id": f"$labels_new.{label}", "count": {"$sum": 1}}}]
        )
        r = [_ for _ in r]
        # r.sort()
        for category_count in r:
            if category_count["_id"] is None:
                continue
            count_all[label][category_count["_id"]] = category_count["count"]

    return count_all


def get_origin_count(db_collection) -> List:
    """Function to return all the number of entries for each unique value of the field "origin" in the given collection

    Args:
        db_collection (pymongo.collection.Collection): pymongo collection

    Returns:
        List: List of tuples containing distinct values of the field "origin" and the number of entries
    """
    r = db_collection.aggregate([{"$group": {"_id": "$origin", "count": {"$sum": 1}}}])
    r = {_["_id"]: _["count"] for _ in r}
    return r


def refresh_label_count(label_list, save_path, db_images):
    all_label_count = get_label_count(label_list, db_images)
    with open(save_path, "w") as f:
        json.dump(all_label_count, f)

    return all_label_count


def refresh_label_by_origin_count(label_list, save_path, db_images):
    label_by_origin_count = get_label_count_by_origin(label_list, db_images)
    with open(save_path, "w") as f:
        json.dump(label_by_origin_count, f)

    return label_by_origin_count


def refresh_videos_origin_count(save_path, db_interventions):
    video_origin_count = get_origin_count(db_interventions)
    with open(save_path, "w") as f:
        json.dump(video_origin_count, f)

    return video_origin_count


def refresh_all(
    db_images,
    db_interventions,
    label_list,
    path_video_origin_count,
    path_label_count,
    path_label_by_origin_count,
):
    label_count = refresh_label_count(label_list, path_label_count, db_images)
    label_by_origin_count = refresh_label_by_origin_count(
        label_list, path_label_by_origin_count, db_images
    )
    video_origin_count = refresh_videos_origin_count(
        path_video_origin_count, db_interventions
    )

    return label_count, label_by_origin_count, video_origin_count



from pathlib import Path
import warnings
from typing import List
from collections import Counter


# Validate if no duplicate video keys exist
def validate_video_keys(db_interventions) -> List:
    """Expects db_intervention collection, filters for non unique video_keys

    Args:
        db_interventions (mongoCollection):

    Returns:
        List: List of non-unique video keys
    """
    interventions = db_interventions.find(
        {"video_key": {"$exists": True}}, {"video_key": 1}
    )
    keys = [_["video_key"] for _ in interventions]
    duplicates = [key for key, count in Counter(keys).items() if count > 1]

    if duplicates:
        warnings.warn("Non unique video keys detected")
    return duplicates


# Validate if files exist
def validate_image_paths(db_images) -> List:
    """
    Expects db image collection. Checks all paths if they exist.
    Warns if any paths do not exist and returns list of image ids where path doesn't exist.
    """
    image_ids = []
    images = db_images.find({}, {"path": 1})
    for _ in images:
        if not Path(_["path"]).exists():
            image_ids.append(_["_id"])
    if image_ids:
        warnings.warn(
            "Not all images for paths of given image collection exist. Returning ID's of invalid images"
        )
    return image_ids


# Validate if video files exist
def validate_video_paths(db_interventions):
    """
    Expects db interventions collection. Checks all entries with have "video_key" if the video file exists.
    Warns if any paths do not exist and returns list of intervention ids where path doesn't exist.
    """
    agg = [{"$match": {"video_path": {"$exists": True}}}]
    videos = db_interventions.aggregate(agg)
    video_ids = []

    for _ in videos:
        if not Path(_["video_path"]).exists():
            video_ids.append(_["_id"])

    if video_ids:
        warnings.warn(
            "Not all videos for paths of given intervention collection exist. Returning ID's of invalid interventions"
        )

    return video_ids



from typing import List
import requests
import xml.etree.ElementTree as ET
import json
from bson.objectid import ObjectId


def delete_project_by_id(project_id: int, labelstudio_prefix: str, labelstudio_token: str):
    requests.delete(f"{labelstudio_prefix}projects/{project_id}", headers={"Authorization": labelstudio_token})


# db -> Labelstudio
def generate_ls_project(label, description, path_label_configs, labelstudio_prefix, labelstudio_token):

    label_config_path = path_label_configs.joinpath(label).with_suffix(".xml")
    assert label_config_path.exists()

    label_config = ET.parse(label_config_path)
    label_config = label_config.getroot()
    label_config = ET.tostring(label_config)

    payload = {
        "title": label,
        "description": description,
        "label_config": label_config,
        "expert_instruction": None,
        "show_instruction": False,
        "show_skip_button": True,
        "enable_empty_annotation": True,
        "show_annotation_history": True,
        "organization": None,
        "color": None,
        "maximum_annotations": 1,
        "is_published": True,
        "model_version": None,
        "sampling": "Uniform sampling",

    }

    r = requests.post(f"{labelstudio_prefix}projects/", headers={"Authorization": labelstudio_token}, data=payload)
    assert r.status_code == 201
    return r.json()


def db_images_to_ls_project(images, label, description, path_label_configs, labelstudio_prefix, labelstudio_token, fileserver_prefix):
    new_project = generate_ls_project(label, description, path_label_configs, labelstudio_prefix, labelstudio_token)
    new_project_id = new_project["id"]
    task_dump = [
        generate_ls_task(image, label, new_project_id, fileserver_prefix) for image in images
        if generate_ls_task(image, label, new_project_id, fileserver_prefix)
        ]
    for task in task_dump:
        task = json.dumps(task)
        r = requests.post(
            f"{labelstudio_prefix}tasks/",
            headers={"Authorization": labelstudio_token, 'content-type': 'application/json'},
            data=task
            )
        assert r.status_code == 201


def generate_ls_task(image, label, new_project_id, fileserver_prefix):
    if "/extreme_storage/files/" in image["path"]:
        task = {
            "data": {
                "image": image["path"].replace("/extreme_storage/files/", fileserver_prefix),
                "prediction": str(image["predictions"][label]["value"]),
                "_id": str(image["_id"])
            },
            "project": str(new_project_id)
        }
        return task


# Labelstudio -> db
def import_ls_project_annotations(project_id, labelstudio_prefix: str, labelstudio_token: str, db_images, verbose: bool = False) -> bool:
    project = requests.get(f"{labelstudio_prefix}projects/{project_id}", headers={"Authorization": labelstudio_token}).json()
    project_labels = project["parsed_label_config"]["choice"]["labels"]

    r = requests.get(
        f"{labelstudio_prefix}projects/{project_id}/tasks/",
        headers={"Authorization": labelstudio_token},
        params={"page_size": "-1"}
        )
    project_tasks = r.json()
    project_tasks_labeled = [_ for _ in project_tasks if _["is_labeled"] is True and _["annotations"]]

    if verbose:
        print(f"Project ID: {project_id}")
        print("Annotated Labels:")
        print(project_labels)
        print(f"Found {len(project_tasks)} Tasks")
        print(f"Processing {len(project_tasks_labeled)} labeled Tasks")
        print("\n")

    with open(f"./ls_dumps/{project_id}.json", "w") as f:
        json.dump(project_tasks, f)
    with open(f"./ls_dumps/{project['id']}_{project['title']}_meta.json", "w") as f:
        json.dump(project, f)

    project_updates = {}
    for task in project_tasks_labeled:
        db_id = task["data"]["_id"]

        project_updates[db_id] = labeled_ls_task_to_db_update(task, project_labels)
    for _id, update in project_updates.items():
        db_images.update_one({"_id": ObjectId(_id)}, {"$set": update})

    return True


def ls_choices_to_label_dict(ls_results: List[dict], label_list: List[str], prefix: str = "labels_new"):
    """Expects labelstudio result (task["annotation"][n_annotation"]["result"]).\
        Asserts that result is of type choices. Reads result and returns label dict.

    Args:
        ls_results (List[dict]): results of ls annotation
        label_list (List[str]): List of all labels to choose from in this project.
        prefix (str, optional): Prefix reffering to field in which the labels are stored\
            in mongo db. Defaults to "labels_new".

    Returns:
        dict: dict of form {prefix.label: bool}
    """
    if ls_results:
        assert len(ls_results) == 1
        result = ls_results[-1]
        assert result["type"] == "choices"
        choices = result["value"]["choices"]
        if "unclear" in choices:
            prefix = "labels_unclear"
        label_dict = {}
        for _ in label_list:
            label_dict[f"{prefix}.{_}"] = False

        for choice in choices:
            label_dict[f"{prefix}.{choice}"] = True
    else:
        label_dict = {}
        for _ in label_list:
            label_dict[f"{prefix}.{_}"] = False

    return label_dict


def labeled_ls_task_to_db_update(task: dict, project_labels: List):
    """Expects a task exported from labelstudio. Task must have at least one annotation result.\
        Lates annotation will be passed to ls_choices_to_label_dict.

    Args:
        task (dict): Labelstudio task with at least one annotation
        project_labels (List): List of all labels available in this project.\
            If not contained in annotation, label will be False.

    Returns:
        dict: Dictionary containig the update.
    """
    assert "annotations" in task
    assert task["annotations"]

    annotations = task["annotations"]
    latest_result = annotations[-1]["result"]
    update_label_dict = ls_choices_to_label_dict(latest_result, project_labels)

    return update_label_dict



from typing import List, Optional

# import re
# import warnings
# from pathlib import Path
# import json


def get_label_types(db_images):
    _labels = db_images.distinct("labels.annotations")
    labels = []
    for _ in _labels:
        labels.extend(list(_.keys()))
    labels = list(set(labels))
    return labels


def get_categorical_label_count(label: str, db_images):
    r = db_images.aggregate(
        [{"$group": {"_id": f"$labels.annotations.{label}", "count": {"$sum": 1}}}]
    )
    r = [_ for _ in r if _["_id"] is not None]
    return r


def get_predictions_without_annotations_query(label: str):
    """
    Expects a label. Returns query dict for images with predictions but not annotations for this label.
    Additionally filters images out if they are already marked as "in_progress".
    !!! Only works for binary classification !!!
    """
    return {
        "$match": {
            f"labels.predictions.{label}": {"$exists": True},
            f"labels.annotation.{label}": {"$exists": False},
        }
    }


def get_ls_task(
    img_path: str,
    img_id: str,
    intervention_id: str,
    targets: List,
    predictions: List,
    origin: str,
    report: Optional[str] = None,
    age: Optional[int] = None,
    gender: Optional[str] = None,
    intervention_date: Optional[str] = None,
    intervention_type: Optional[str] = None,
) -> dict:
    """Function to generate a labelstudio task element

    Args:
        img_path (str): Path to load image in labelstudio. In most cases this should be a URL like 'http://localhost:8000/{...}'
            or an absolute path to the image.
        img_id (str): String of the MongoDb ObjectId of the image to label
        targets (List): List of targets to label
        report (str): String of the report associated with Image, can be empty.
        origin (str): Source of the image.
        age (int): Age of the individual the image was taken from at the time of the intervention
        gender (str): Biological gender of the individual the image was taken from.
        intervention_date (str): Date string in format "%Y-%m-%d" specifying the intervention date.
        intervention_id (str): String of the MongoDb ObjectId of the image's intervention.
        intervention_type (str): Intervention Type, e.g. "Koloskopie".
        predictions (List): List of the image's predictions.

    Returns:
        dict: Returns dictionary of a single labelstudio task with the keys "data"
        (contains: "image", "_id", "report", "origin", "age", "gender", "intervention_date", "intervention_id", "intervention_type",
        "targets) and "predictions"
    """
    _dict = {
        "data": {
            "image": img_path,
            "_id": img_id,
            "report": report,
            "origin": origin,
            "age": age,
            "gender": gender,
            "intervention_date": intervention_date,
            "intervention_id": intervention_id,
            "intervention_type": intervention_type,
            "targets": targets,
        },
        "predictions": predictions,
    }
    delete = []
    if not _dict["predictions"]:
        del _dict["predictions"]
    for key, value in _dict["data"].items():
        if value is None:
            delete.append(key)
    for key in delete:
        del _dict["data"][key]

    return _dict


def db_to_ls_classification_labels(db_label: dict) -> dict:
    """Function to convert database classification labels to labelstudio labels.

    Args:
        db_label (dict): Dictionary contains labels as keys with associated bool values.

    Returns:
        dict: Returns dictionary with key "choices" and a list of all labels which were evaluated as true.
    """
    ls_label = []
    for target, value in db_label.items():
        if value == 1:
            ls_label.append(target)

    return {"choices": ls_label}


def ls_to_db_labels(ls_label) -> dict:
    """Fuction to convert labelstudio classification annotations to database compatible label dictionaries

    Args:
        ls_label (dict): Expects a labelstudio task element with an annotation

    Returns:
        dict: database label element
    """
    db_label = {}
    db_label["_id"] = ls_label["data"]["_id"]
    db_label["targets"] = ls_label["data"]["targets"]
    db_label["annotations"] = [_["result"] for _ in ls_label["annotations"]]
    if db_label["annotations"]:
        assert len(db_label["annotations"]) == 1
        db_label["annotations"] = db_label["annotations"][0]
        if db_label["annotations"]:
            assert len(db_label["annotations"]) == 1
            db_label["annotations"] = db_label["annotations"][0]
            db_label["labels"] = {_: False for _ in db_label["targets"]}

            for key, value in db_label["annotations"]["value"].items():
                # Add Transformation for boxes ###################
                if key == "choices":
                    for _label in value:
                        db_label["labels"][_label] = True

            return db_label


# Patterns to parse first round annotations
# regex_patterns = {
#     "anatomy_ileocecalvalve": re.compile("anatomie.ileocecalvalve|ileocecal\svalve"),
#     "anatomy_ileum": re.compile("(.*)\.*(ileum)"),
#     "anatomy_appendix": re.compile("appendix"),
#     "polyp": re.compile("p\d\.*(.*)"),
#     "polyp_nice": re.compile("nice(.*)"),
#     "normal": re.compile("normal\.(.*)|normalmucosa\.(.*)"),
#     "polyp_paris": re.compile("paris(.*)"),
#     "polyp_size": re.compile("size(.*)"),
#     "igien": re.compile("igien"),
#     "diverticula": re.compile("diverticula"),
# }


# def process_webserver_json(
#     video_json: dict, base_path_frames: Path, skip_frame_factor: int, path_logs: Path
# ) -> dict:
#     """Processes single json from webserver. Returns intervention insert dict and list of frame insert dicts.

#     Args:
#         video_json (dict): json from webserver
#         base_path_frames (Path): Path object pointing to directory where all frame paths are stored
#         skip_frame_factor (int): every n-th frame was annotated in first round
#         path_logs (Path): Path object pointing to directory where logs should be stored

#     Returns:
#         dict: [description]
#     """
#     video_path = Path(video_json["videoPath"])
#     video_hash = video_path.parent.name
#     annotation_paths = [Path(_) for _ in video_json["annotationPaths"]]
#     intervention_type = get_intervention_type(video_json["videoType"])

#     annotations = []
#     for _ in annotation_paths:
#         with open(_, "r", encoding="latin1") as f:
#             annotations.append(json.load(f))

#     # Select only latest annotation
#     annotation = annotations[-1]

#     # Validate
#     if not validate_adrian_annotation_json(annotation):
#         return None
#     else:
#         intervention_content = get_intervention_content(
#             annotation, video_path, intervention_type, skip_frame_factor
#         )
#         intervention_content["origin"] = video_json["center"].lower()
#         insert_intervention, insert_frames = make_insert_dicts(
#             intervention_content, base_path_frames
#         )

#         insert_intervention["video_hash"] = video_hash
#         if "patho" in video_json:
#             insert_intervention["patho_raw"] = video_json["patho"]
#         if "report" in video_json:
#             insert_intervention["report_raw"] = video_json["report"]

#         if Path(insert_frames[0]["path"]).parent.exists():
#             with open(
#                 path_logs.joinpath(f"{Path(video_json['videoPath']).name}.json"), "w"
#             ) as f:
#                 json.dump(video_json, f)
#             warnings.warn(f"Path already exists!\n{video_path}")
#             return None
#         else:
#             # if "frames" in insert_intervention:
#             #     insert_intervention["frames"] = [str(_) for _ in insert_intervention["frames"]]

#             return {
#                 "insert_intervention": insert_intervention,
#                 "insert_frames": insert_frames,
#             }


# def get_intervention_content(
#     annotation: dict, video_path: Path, intervention_type: str, skip_frame_factor: int
# ):
#     """Create an intervention_content dictionary containing.

#     Args:
#         annotation (dict): Adrian annotation file
#         video_path (Path): Path object pointing to video file
#         intervention_type (str): type of intervention
#         skip_frame_factor (int): Indicates how many frames were initially skipped for first round annotation, e.g.:
#         3 means that every third frame was extracted

#     Returns:
#         dict: Dictionary containing
#         "video_key", "video_path", "annotation_raw", "frames_total", "annotation", "origin", "intervention_type"
#     """
#     intervention_content = {}

#     extracted_annotation = annotation_conversion(annotation, skip_frame_factor)
#     key = extracted_annotation["video_key"]

#     intervention_content["video_key"] = key
#     intervention_content["video_path"] = video_path.as_posix()
#     intervention_content["annotation_raw"] = annotation
#     intervention_content["frames_total"] = annotation["metadata"]["imageCount"]
#     intervention_content["annotation"] = extracted_annotation
#     intervention_content["intervention_type"] = intervention_type

#     return intervention_content


# def match_labels(label, regex_patterns):
#     for key, pattern in regex_patterns.items():
#         _ = pattern.match(label)
#         if _:
#             if key == "polyp_nice":
#                 _ = {"polyp_nice": _.groups()[0]}
#             if key == "polyp_paris":
#                 _ = {"polyp_paris": _.groups()[0]}
#             if key == "igien":
#                 _ = {"igien": True}
#             if key == "polyp_size":
#                 _ = {"polyp_size": _.groups()[0]}
#             if key == "polyp":
#                 if _.groups()[0]:
#                     _ = {"polyp": True, "type": _.groups()[0]}
#                 else:
#                     _ = {"polyp": True, "type": "single"}
#             if key == "normal":
#                 _type = None
#                 for _group in _.groups():
#                     if _group:
#                         if _group == "start":
#                             _type = "n_start"
#                         elif "end" in _group:
#                             _type = "n_ende"
#                 if not _type:
#                     _type = "single"
#                 _ = {"polyp": False, "type": _type}
#             if key == "anatomy_ileocecalvalve":
#                 _ = {"anatomy": "ileocecalvalve"}
#             if key == "anatomy_ileum":
#                 if _.groups()[0]:
#                     _ = {"anatomy": {"ileum": _.groups()[0].replace(".", "")}}
#                 else:
#                     _ = {"anatomy": "ileum"}

#             if key == "anatomy_appendix":
#                 _ = {"anatomy": "appendix"}
#             if key == "diverticula":
#                 _ = {"diverticulum": True}

#             return _


# def get_intervention_type(_type):
#     _type = _type.strip().lower()
#     if _type == "kolo":
#         _type = "Koloskopie"
#     elif _type == "coloscopy":
#         _type = "Koloskopie"
#     elif _type == "gastro":
#         _type = "Gastroskopie"
#     elif _type == "ögd":
#         _type = "Gastroskopie"
#     elif _type == "gastro/kolo":
#         _type = "Gastroskopie/Koloskopie"
#     elif _type == "ögd/kolo":
#         _type = "Gastroskopie/Koloskopie"
#     elif _type == "kolo0":
#         _type = "Koloskopie"
#     else:
#         warnings.warn(f"unknown type:{_type}")

#     return _type


# def extract_superframes(superframes) -> dict:
#     annotations = []
#     frame_numbers = []

#     for frame in superframes["childFrames"]:
#         label = [match_labels(_.lower(), regex_patterns) for _ in frame["classes"]]
#         annotations.append(label)
#         frame_numbers.append(frame["frameNumber"])

#     labels = {_: [] for _ in frame_numbers}
#     for i, label in enumerate(annotations):
#         for _class in label:
#             labels[frame_numbers[i]].append(_class)

#     return labels


# def extract_box_images(box_images):
#     annotations = {}
#     for image in box_images:
#         bbs = []
#         for _bb in image["boundingBox"]:
#             if "secondaryLabels" in _bb:
#                 sec_labels = _bb["secondaryLabels"].copy()
#             else:
#                 sec_labels = []
#             bb = {
#                 "x": _bb["x"],
#                 "y": _bb["y"],
#                 "width": _bb["width"],
#                 "height": _bb["height"],
#                 "labels": sec_labels,
#             }
#             bb["labels"].append(_bb["label"])
#             bb["labels"] = [
#                 match_labels(_.lower(), regex_patterns) for _ in bb["labels"]
#             ]
#             bbs.append(bb)
#         annotations[image["frameNumber"]] = bbs
#         if image["label"]:
#             _ = match_labels(image["label"].lower(), regex_patterns)
#             annotations[image["frameNumber"]].append(_)

#     return annotations


# def fill_polyp_frames(video_object):
#     state = None
#     updates = []
#     start_frame = None
#     n_start_frame = None
#     for frame_number, annotation_list in video_object["classifications"].items():
#         for annotation in annotation_list:
#             if annotation:
#                 if "polyp" in annotation:
#                     if annotation["polyp"] is True:
#                         state = annotation["type"]

#                         if state == "start":
#                             start_frame = frame_number

#                         elif state == "single":
#                             pass
#                         elif state == "end":
#                             if start_frame:
#                                 updates.extend(
#                                     [
#                                         {_: {"polyp": True, "type": "filled"}}
#                                         for _ in range(start_frame, frame_number + 1)
#                                     ]
#                                 )
#                                 start_frame = None
#                             else:
#                                 warnings.warn(
#                                     f"End frame of annotation recognized before start frame. \
#                                         This annotation will be omitted: \n{annotation}"
#                                 )

#                     if annotation["polyp"] is False:
#                         n_state = annotation["type"]
#                         if n_state == "n_start":
#                             n_start_frame = frame_number
#                         elif n_state == "single":
#                             pass
#                         elif n_state == "n_ende":
#                             if n_start_frame:
#                                 updates.extend(
#                                     [
#                                         {_: {"polyp": False, "type": "filled"}}
#                                         for _ in range(n_start_frame, frame_number + 1)
#                                     ]
#                                 )
#                                 n_start_frame = None
#                             else:
#                                 warnings.warn(
#                                     f"End frame of annotation recognized before start frame. \
#                                         This annotation will be omitted: \n{annotation}"
#                                 )

#     for update in updates:
#         if update:
#             _frame_number = list(update.keys())[0]
#             if _frame_number in video_object["classifications"]:
#                 video_object["classifications"][_frame_number].append(
#                     update[_frame_number]
#                 )
#             else:
#                 video_object["classifications"][_frame_number] = [update[_frame_number]]

#     return video_object


# def fill_ileum_frames(video_object):
#     state = None
#     updates = []
#     for frame_number, annotation_list in video_object["classifications"].items():
#         for annotation in annotation_list:
#             if annotation:
#                 if "anatomy" in annotation:
#                     if isinstance(annotation["anatomy"], dict):
#                         if "ileum" in annotation["anatomy"]:
#                             state = annotation["anatomy"]["ileum"]
#                             if state == "enter":
#                                 start_frame = frame_number
#                             elif state == "single":
#                                 pass
#                             elif state == "exit":
#                                 updates.extend(
#                                     [
#                                         {_: {"anatomy": "ileum"}}
#                                         for _ in range(start_frame, frame_number + 1)
#                                     ]
#                                 )

#     for update in updates:
#         if update:
#             _frame_number = list(update.keys())[0]
#             if _frame_number in video_object["classifications"]:
#                 video_object["classifications"][_frame_number].append(
#                     update[_frame_number]
#                 )
#             else:
#                 video_object["classifications"][_frame_number] = [update[_frame_number]]

#     return video_object


# def multiply_frame_keys(frame_dict: dict, skip_frame_factor: int):
#     _classifications = {}
#     for _key, value in frame_dict.items():
#         new_key = _key * skip_frame_factor
#         _classifications[new_key] = value
#     return _classifications


# def annotation_conversion(annotation: dict, skip_image_factor: int):
#     video_object = {
#         "video_key": annotation["metadata"]["videoFile"],
#         "width": annotation["metadata"]["resolutionWidth"],
#         "height": annotation["metadata"]["resolutionHeight"],
#         "image_count": annotation["metadata"]["imageCount"] * skip_image_factor,
#         "classifications": [
#             extract_superframes(superframe)
#             for superframe in annotation["superframes"]
#             if extract_superframes(superframe)
#         ],
#         "bounding_boxes": extract_box_images(annotation["images"]),
#     }
#     _tmp = {}
#     for _ in video_object["classifications"]:
#         for key, value in _.items():
#             if key in _tmp:
#                 _tmp[key].extend(value)
#             else:
#                 _tmp[key] = value

#     video_object["classifications"] = _tmp

#     video_object["classifications"] = multiply_frame_keys(
#         video_object["classifications"], skip_image_factor
#     )
#     video_object["bounding_boxes"] = multiply_frame_keys(
#         video_object["bounding_boxes"], skip_image_factor
#     )

#     video_object = fill_polyp_frames(video_object)
#     video_object = fill_ileum_frames(video_object)
#     return video_object


# def make_insert_dicts(intervention: dict, base_path_frames: Path):
#     frames = []
#     frames.extend(list(intervention["annotation"]["classifications"].keys()))
#     frames.extend(list(intervention["annotation"]["bounding_boxes"].keys()))
#     frames = list(set(frames))
#     frames.sort()

#     # intervention["frames"] = [str(_) for _ in frames]

#     image_dicts = []
#     for frame in frames:
#         _dict = {
#             "origin": intervention["origin"],
#             "video_key": intervention["video_key"],
#             "intervention_id": None,
#             "path": base_path_frames.joinpath(intervention["video_key"])
#             .joinpath(f"{frame}.png")
#             .as_posix(),
#             "labels": {"annotations": {}, "predictions": {}},
#             "n_frame": frame,
#             "in_progress": False,
#         }

#         _annotations = []
#         if frame in intervention["annotation"]["bounding_boxes"]:
#             bb_labels = []
#             for bbox in intervention["annotation"]["bounding_boxes"][frame]:
#                 bbox["labels"] = [_ for _ in bbox["labels"] if _]
#                 for _bb_label in bbox["labels"]:
#                     for key, value in _bb_label.items():
#                         bb_labels.append({key: value})
#                 bb_labels.append(
#                     {
#                         "polyp_detection_bbox": {
#                             key: value for key, value in bbox.items() if key != "labels"
#                         }
#                     }
#                 )
#             _annotations.extend(bb_labels)

#         if frame in intervention["annotation"]["classifications"]:
#             _annotations.extend(intervention["annotation"]["classifications"][frame])

#         annotations = {}
#         for _ in _annotations:
#             if _:
#                 key = list(_.keys())[0]
#                 if key in annotations:
#                     pass
#                 elif key == "type":
#                     pass
#                 else:
#                     annotations[key] = _[key]

#         _dict["labels"]["annotations"] = annotations
#         # del intervention["annotation"]
#         image_dicts.append(_dict)

#     return intervention, image_dicts


# def validate_adrian_annotation_json(annotation: dict) -> bool:
#     """Function to validate json files from adrians annotation tool. Returns
#     False if any of "metadata", "metadata.videofile", "superframes" or "images"
#     are missing.

#     Args:
#         annotation (dict): Dictionarie of read annotation json.

#     Returns:
#         bool: True if valid, else false.
#     """
#     if "metadata" not in annotation:
#         warnings.warn("Annotation does not contain Metadata")
#         return False
#     if "images" not in annotation:
#         warnings.warn("Annotation does not contain Image-Array")
#         return False
#     if "superframes" not in annotation:
#         warnings.warn("Annotation does not contain Superframe-Array")
#         return False
#     if "videoFile" not in annotation["metadata"]:
#         warnings.warn("Annotation Metadata does not contain a video file name")
#         return False
#     if "imageCount" not in annotation["metadata"]:
#         warnings.warn("Annotation Metadata does not contain an image count")
#     return True
